from .environment import Env
from anndata import AnnData
import torch
import tqdm

class agent(Env):
    """  
    An agent class for modeling single-cell transcriptomics data using a variational autoencoder approach.  

    Parameters  
    ----------  
    adata : AnnData  
        Annotated data matrix.  
    layer : str, optional  
        The layer of the AnnData object to use, by default 'counts'.  
    percent : float, optional  
        Percent parameter value, by default 0.01.  
    irecon : float, optional  
        Irecon parameter value, by default 0.0.  
    beta : float, optional  
        Beta parameter value, by default 1.0.  
    dip : float, optional  
        Dip parameter value, by default 0.0.  
    tc : float, optional  
        TC parameter value, by default 0.0.  
    info : float, optional  
        Info parameter value, by default 0.0.  
    hidden_dim : int, optional  
        Hidden dimension size, by default 128.  
    latent_dim : int, optional  
        Latent dimension size, by default 10.  
    i_dim : int, optional  
        I dimension size, by default 2.  
    lr : float, optional  
        Learning rate, by default 1e-4.  
    device : torch.device, optional  
        Device to run the model on, by default uses GPU if available, otherwise CPU.  

    Methods  
    -------  
    fit(epochs=1000)  
        Fits the model to the data for a specified number of epochs.  
    get_iembed()  
        Returns the intermediate embedding from the neural network.  
    get_latent()  
        Returns the latent representation of the data.  
    """ 
    def __init__(
        self,
        adata: AnnData,
        layer: str = 'counts',
        percent: float = .01,
        irecon: float = .0,
        beta: float = 1.,
        dip: float = .0,
        tc: float = .0,
        info: float = .0,
        hidden_dim: int = 128,
        latent_dim: int = 10,
        i_dim: int = 2,
        lr: float = 1e-4,
        device: torch.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    ):
        super().__init__(
            adata=adata,
            layer=layer,
            percent=percent,
            irecon=irecon,
            beta=beta,
            dip=dip,
            tc=tc,
            info=info,
            hidden_dim=hidden_dim,
            latent_dim=latent_dim,
            i_dim=i_dim,
            lr=lr,
            device=device
        )
        
    def fit(
        self,
        epochs:int=1000
    ):
        """  
        Fits the model to the data for a specified number of epochs.  

        Parameters  
        ----------  
        epochs : int, optional  
            Number of training epochs, by default 1000.  

        Returns  
        -------  
        agent  
            The fitted agent instance.  
        """
        with tqdm.tqdm(total=int(epochs), desc='Fitting', ncols=150) as pbar:
            for i in range(int(epochs)):
                data = self.load_data()
                self.step(data)
                if (i+1) % 10 == 0:
                    pbar.set_postfix({
                        'Loss':f'{self.loss[-1][0]:.2f}',
                        'ARI':f'{(self.score[-1][0]):.2f}',
                        'NMI':f'{(self.score[-1][1]):.2f}',
                        'ASW':f'{(self.score[-1][2]):.2f}',
                        'C_H':f'{(self.score[-1][3]):.2f}',
                        'D_B':f'{(self.score[-1][4]):.2f}',
                        'P_C':f'{(self.score[-1][5]):.2f}'
                    })
                pbar.update(1)
        return self
    
    def get_iembed(
        self
    ):
        """  
        Returns the intermediate embedding from the neural network.  

        Returns  
        -------  
        numpy.ndarray  
            The intermediate embedding as a NumPy array.  
        """
        return self.nn(torch.tensor(self.X, dtype=torch.float).to(self.device))[-3].detach().cpu().numpy()    
        
    def get_latent(
        self,
    ):
        """  
        Returns the latent representation of the data.  

        Returns  
        -------  
        numpy.ndarray  
            The latent representation as a NumPy array.  
        """
        q_z = self.take_latent(self.X)
        return q_z



import numpy as np
from numpy import ndarray
import pandas as pd
import scib
from sklearn.cluster import KMeans
from sklearn.neighbors import kneighbors_graph
from sklearn.metrics import adjusted_mutual_info_score, normalized_mutual_info_score, silhouette_score, calinski_harabasz_score, davies_bouldin_score
from scipy.sparse import csr_matrix 
from scipy.sparse import csgraph

def get_dfs(
    mode, 
    agent_list
):
    if mode == 'mean':
        ls = list(map(lambda x: zip(*(np.array(b).mean(axis=0) for b in zip(*((zip(*a.score)) for a in x)))), list(zip(*agent_list))))
    else:
        ls = list(map(lambda x: zip(*(np.array(b).std(axis=0) for b in zip(*((zip(*a.score)) for a in x)))), list(zip(*agent_list))))
    return (map(lambda x:pd.DataFrame(x, columns=['ARI', 'NMI', 'ASW', 'C_H', 'D_B', 'P_C']),ls))

def moving_average(
    a, 
    window_size
):
    cumulative_sum = np.cumsum(np.insert(a, 0, 0)) 
    middle = (cumulative_sum[window_size:] - cumulative_sum[:-window_size]) / window_size
    r = np.arange(1, window_size-1, 2)
    begin = np.cumsum(a[:window_size-1])[::2] / r
    end = (np.cumsum(a[:-window_size:-1])[::2] / r)[::-1]
    return np.concatenate((begin, middle, end))

def fetch_score(
    adata, 
    latent, 
    label_true, 
    label_mode='KMeans', 
    batch=False
):
    q_z = latent
    if label_mode == 'KMeans':
        labels = KMeans(q_z.shape[1]).fit_predict(q_z)
    elif label_mode == 'Max': 
        labels = np.argmax(q_z, axis=1)
    elif label_mode == 'Min':
        labels = np.argmin(q_z, axis=1)
    else:
        raise ValueError('Mode must be in one of KMeans, Max and Min')
        
    adata.obsm['X_qz'] = q_z
    adata.obs['label'] = pd.Categorical(labels)
    
    NMI = normalized_mutual_info_score(label_true, labels)
    ARI = adjusted_mutual_info_score(label_true, labels)
    ASW = silhouette_score(q_z, labels) 
    if label_mode != 'KMeans':
        ASW = abs(ASW)
    C_H = calinski_harabasz_score(q_z, labels)
    D_B = davies_bouldin_score(q_z, labels)
    G_C = graph_connection(kneighbors_graph(adata.obsm['X_qz'], 15), adata.obs['label'].values)
    clisi = scib.metrics.clisi_graph(adata, 'label', 'embed', 'X_qz', n_cores=26)

    if batch:
        sub_adata = adata[np.random.choice(adata.obs_names, 5000, replace=False)].copy()
        ilisi = scib.metrics.ilisi_graph(sub_adata, 'batch', 'embed', 'X_qz', n_cores=26)
        bASW = scib.metrics.silhouette_batch(sub_adata, 'batch', 'label', 'X_qz')
        print('Completed')
        return NMI, ARI, ASW, C_H, D_B, G_C, clisi, ilisi, bASW
    print('Completed')
    return NMI, ARI, ASW, C_H, D_B, G_C, clisi

def graph_connection(
    graph: csr_matrix,
    labels: ndarray
):
    cg_res = []
    for l in np.unique(labels):
        mask = np.where(labels==l)[0]
        subgraph = graph[mask, :][:, mask]
        _, lab = csgraph.connected_components(subgraph, connection='strong')
        tab = np.unique(lab, return_counts=True)[1]
        cg_res.append(tab.max() / tab.sum())
    return np.mean(cg_res)







import torch
import torch.nn as nn
import torch.optim as optim
from .mixin import scviMixin, dipMixin, betatcMixin, infoMixin
from .module import VAE

class iVAE(scviMixin, dipMixin, betatcMixin, infoMixin):
    def __init__(
        self,
        irecon,
        beta,
        dip,
        tc,
        info,
        state_dim, 
        hidden_dim, 
        latent_dim,
        i_dim,
        lr,
        device,
        *args, 
        **kwargs
    ):
        self.irecon = irecon
        self.beta = beta
        self.dip = dip
        self.tc = tc
        self.info = info
        self.nn = VAE(state_dim, hidden_dim, latent_dim, i_dim).to(device)
        self.nn_optimizer = optim.Adam(self.nn.parameters(), lr=lr)
        self.device = device
        self.loss = []
    
    def take_latent(
        self, 
        state
    ):
        state = torch.tensor(state, dtype=torch.float).to(self.device)
        q_z, _, _, _, _, _, _ = self.nn(state)
        return q_z.detach().cpu().numpy()
        
    def update(
        self, 
        states
    ):
        states = torch.tensor(states, dtype=torch.float).to(self.device)
        q_z, q_m, q_s, pred_x, le, ld, pred_xl = self.nn(states)
       
        l = states.sum(-1).view(-1,1)
        pred_x = pred_x * l
        
        disp = torch.exp(self.nn.decoder.disp)
        recon_loss = -self._log_nb(states, pred_x, disp).sum(-1).mean()
        
        if self.irecon:
            pred_xl = pred_xl * l
            irecon_loss = - self.irecon * self._log_nb(states, pred_xl, disp).sum(-1).mean()
        else:
            irecon_loss = torch.zeros(1).to(self.device)
        
        p_m = torch.zeros_like(q_m)
        p_s = torch.zeros_like(q_s)
        
        kl_div = self.beta * self._normal_kl(q_m, q_s, p_m, p_s).sum(-1).mean()
        
        if self.dip:
            dip_loss = self.dip * self._dip_loss(q_m ,q_s)
        else:
            dip_loss = torch.zeros(1).to(self.device)
        
        if self.tc:
            tc_loss = self.tc * self._betatc_compute_total_correlation(q_z, q_m ,q_s)
        else:
            tc_loss = torch.zeros(1).to(self.device)
        
        if self.info:
            mmd_loss = self.info * self._compute_mmd(q_z, torch.randn_like(q_z))
        else:
            mmd_loss = torch.zeros(1).to(self.device)
        
        total_loss = recon_loss + irecon_loss + kl_div + dip_loss + tc_loss + mmd_loss
            
        self.nn_optimizer.zero_grad()
        total_loss.backward()
        self.nn_optimizer.step()

        self.loss.append((
            total_loss.item(),
            recon_loss.item(), 
            irecon_loss.item(),
            kl_div.item(), 
            dip_loss.item(),
            tc_loss.item(), 
            mmd_loss.item()
        ))


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Normal

def weight_init(m):
    if isinstance(m, nn.Linear):
        nn.init.xavier_normal_(m.weight)
        nn.init.constant_(m.bias, .01)

class Encoder(nn.Module):
    def __init__(
        self, 
        state_dim, 
        hidden_dim, 
        action_dim
    ):
        super(
            Encoder, 
            self
        ).__init__()
        self.nn = nn.Sequential(
            nn.Linear(state_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, action_dim*2)
        )
        self.apply(weight_init)

    def forward(self, x):
        output = self.nn(x)
        q_m = output[:,:int(output.shape[-1]/2)]
        q_s = output[:,int(output.shape[-1]/2):]     
        s = F.softplus(q_s) + 1e-6
        n = Normal(q_m, s)
        q_z = n.rsample()
        return q_z, q_m, q_s


class Decoder(nn.Module):
    def __init__(
        self, 
        state_dim, 
        hidden_dim, 
        action_dim
    ):
        super(
            Decoder, 
            self
        ).__init__()
        self.nn = nn.Sequential(
            nn.Linear(action_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, state_dim),
            nn.Softmax(dim=-1)
        )
        self.disp = nn.Parameter(torch.randn(state_dim))
        self.apply(weight_init)

    def forward(self, x):
        output = self.nn(x)
        return output
        
class VAE(nn.Module):
    def __init__(
        self, 
        state_dim, 
        hidden_dim, 
        action_dim,
        i_dim
    ):
        super(
            VAE, 
            self
        ).__init__()
        self.encoder = Encoder(state_dim, hidden_dim, action_dim)
        self.decoder = Decoder(state_dim, hidden_dim, action_dim)
        self.latent_encoder = nn.Linear(action_dim, i_dim)
        self.latent_decoder = nn.Linear(i_dim, action_dim)
        
    def forward(
        self, 
        x
    ):
        
        q_z, q_m, q_s = self.encoder(x)
        
        le = self.latent_encoder(q_z)
        ld = self.latent_decoder(le)
        
        pred_x = self.decoder(q_z)
        pred_xl = self.decoder(ld)
        
        return q_z, q_m, q_s, pred_x, le, ld, pred_xl


#!/usr/bin/env python

import argparse
import os
from .agent import agent
import torch
from anndata import read_h5ad
import numpy as np

def main():
    parser = argparse.ArgumentParser(description='Run the agent class training script')
    parser.add_argument('--epochs', type=int, default=1000, help='Number of training epochs')
    parser.add_argument('--layer', type=str, default='counts', help='Layer to use from adata')
    parser.add_argument('--percent', type=float, default=0.01, help='Percent parameter value')
    parser.add_argument('--irecon', type=float, default=0.0, help='Irecon parameter value')
    parser.add_argument('--beta', type=float, default=1.0, help='Beta parameter value')
    parser.add_argument('--dip', type=float, default=0.0, help='Dip parameter value')
    parser.add_argument('--tc', type=float, default=0.0, help='TC parameter value')
    parser.add_argument('--info', type=float, default=0.0, help='Info parameter value')
    parser.add_argument('--hidden_dim', type=int, default=128, help='Hidden dimension size')
    parser.add_argument('--latent_dim', type=int, default=10, help='Latent dimension size')
    parser.add_argument('--i_dim', type=int, default=2, help='i dimension size')
    parser.add_argument('--lr', type=float, default=1e-4, help='Learning rate')
    parser.add_argument('--data_path', type=str, default='data.h5ad', help='Path to the data file')
    parser.add_argument('--output_dir', type=str, default='iVAE_output', help='Directory to save the results')

    args = parser.parse_args()

    # Load data
    adata = read_h5ad(args.data_path)

    # Initialize the agent object
    ag = agent(
        adata=adata,
        layer=args.layer,
        percent=args.percent,
        irecon=args.irecon,
        beta=args.beta,
        dip=args.dip,
        tc=args.tc,
        info=args.info,
        hidden_dim=args.hidden_dim,
        latent_dim=args.latent_dim,
        i_dim=args.i_dim,
        lr=args.lr,
        device=args.device if hasattr(args, 'device') else torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    )

    # Train the model
    ag.fit(epochs=args.epochs)

    # Get the latent space representations
    iembed = ag.get_iembed()
    latent = ag.get_latent()

    # Create the output directory if it doesn't exist
    os.makedirs(args.output_dir, exist_ok=True)

    # Save the results
    np.save(os.path.join(args.output_dir, 'iembed.npy'), iembed)
    np.save(os.path.join(args.output_dir, 'latent.npy'), latent)

    print(f"Results have been saved to the '{args.output_dir}' directory.")

if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name = 'iVAE',
    version = '0.0.3',
    description = 'Interpretable Variational Autoencoder for Single-Cell Data',
    long_description = open('README.md').read(),
    long_description_content_type='text/markdown',
    url = 'https://github.com/PeterPonyu/iVAE',
    author = 'Zeyu Fu',
    author_email = 'fuzeyu99@126.com',
    license = 'MIT',
    packages = find_packages(),
    install_requires = [
        'scanpy>=1.10.4',
        'scib>=1.1.5',
        'torch>=1.13.1',
        'joblib>=1.2.0',
        'tqdm>=4.64.1',
    ],
    entry_points={
        'console_scripts': [
            'iVAE=iVAE.cli:main',
        ]
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    python_requires='>=3.9',
)



from .model import iVAE
from .mixin import envMixin
import numpy as np
from sklearn.cluster import KMeans


class Env(iVAE, envMixin):
    
    def __init__(
        self,
        adata,
        layer,
        percent,
        irecon,
        beta,
        dip,
        tc,
        info,
        hidden_dim,
        latent_dim,
        i_dim,
        lr,
        device,
        *args,
        **kwargs
    ):
        self._register_anndata(adata, layer, latent_dim)
        self.batch_size = int(percent * self.n_obs)
        super().__init__(
            irecon     = irecon,
            beta       = beta,
            dip        = dip,
            tc         = tc,
            info       = info,
            state_dim  = self.n_var,
            hidden_dim = hidden_dim, 
            latent_dim = latent_dim,
            i_dim      = i_dim,
            lr         = lr, 
            device     = device
        )
        self.score = []
    
    def load_data(
        self,
    ):
        data, idx = self._sample_data()
        self.idx = idx
        return data
        
    def step(
        self,
        data
    ):
        self.update(data)
        latent = self.take_latent(data)
        score = self._calc_score(latent)
        self.score.append(score)
    
    def _sample_data(
        self,
        
    ):
        idx = np.random.permutation(self.n_obs)
        idx_ = np.random.choice(idx, self.batch_size)
        data = self.X[idx_,:]
        return data, idx_

    def _register_anndata(
        self,
        adata,
        layer: str,
        latent_dim
    ):
        self.X = np.log1p(adata.layers[layer].A)
        self.n_obs = adata.shape[0]
        self.n_var = adata.shape[1]
        self.labels = KMeans(latent_dim).fit_predict(self.X)
        return 